poem = """
Roses are red, violets are blue,
naming your package is really hard to undo.
Haste can make decisions in one fell swoop,
note that LEGO® is a trademark of the LEGO Group.
It really makes sense, we do not need to bluff,
LEGO does not sponsor, authorize or endorse any of this stuff.

Look at all the features and look at all the extensions,
the path towards ruin is paved with good intentions.
Be careful with features as they tend to go sour,
defer responsibility to the end user, this might just give them power.
If you don't know the requirements you don't know if they're met.
If you haven't gotten to where you're going, you aren't there yet.

Infinity is ever present, the unknown may be ignored,
not everything needs to be built, not everything needs to be explored.
Change everything and you'll soon be a jerk,
you may invent a new tool, not a way to work.
Some problems cannot be solved in a single day,
but if you ignore them, they sometimes go away.

There's a lot of power in simplicity,
it keeps you approach strong,
if you understand the solution better than the problem,
you're doing it wrong.
"""

print(poem)
